#include <iostream>
#include <fstream>
#include <string>
#include <ctype.h>
#include <assert.h>
#include <windows.h>
using namespace std;


class RefStr;

class RefStrBuf
{
   char   *StrPtr;                               // -> beginning of string.  May be NULL if       //
                                                 // string buffer does not currently store a      //
                                                 // string.                                       //
   int     UseCnt;                               // String buffer usage count.  Indicates the     //
                                                 // number of string objects that use the string  //
                                                 // buffer.  This count should always be greater  //
                                                 // than 0.  When the count goes to 0, the string //
                                                 // buffer will delete itself.                    //
   int     Len;                                  // Current string length.  The length may be 0.  //
                                                 // The length must not exceed the size of the    //
                                                 // buffer specified by the Siz member.           //
   int     Siz;                                  // Current buffer storage space size.  The size  //
                                                 // may be 0.  The size must be greater than or   //
                                                 // equal to the length specified by the Len      //
                                                 // member.                                       //
   friend  RefStr;
public:
   void Chk() const;
   RefStrBuf();
   RefStrBuf(const char *Str,int StrLen = -1,int StrSiz = -1);
   ~RefStrBuf();
   void Inc();
   void Dec();
   void ReSiz(int NewLen,int NewSiz);
private:
   RefStrBuf(const RefStrBuf &Src);
   RefStrBuf &operator =(const RefStrBuf &Rht);
};
////
class RefStr
{
   RefStrBuf *BufPtr;                            // -> string buffer storing the string.  This    //
                                                 // may be NULL if the string object stores no    //
                                                 // string.                                       //
public:
   void Chk() const;
   RefStr();
   RefStr(const RefStr &Src);
   RefStr(const char *Str,int StrLen = -1);
   ~RefStr();
   int length(void) const;
   RefStr &operator = (const RefStr &Src);
   RefStr &operator += (const RefStr &Str);
   RefStr &operator += (char  Chr);
   const char &operator [] (int Ndx) const;
   char &operator [] (int Ndx);
   char *GetPtr(int Ndx);
   const char *GetStrPtr(int Ndx) const;
   const char *data() const;
   char *GetStrPtr(int Ndx);
   void MakSubStr(int SubOff,int SubLen);
   RefStr operator () (int SubOff,int SubLen) const;
   RefStr substr (int SubOff,int SubLen) const;
   void RTrm();
   RefStr RTrim() const;
   void LTrm();
   RefStr LTrim() const;
   void Trm();
   RefStr Trim() const;
   void PreWrt(int NewSiz,bool CpyFlg);
};
////
class OptStr;

class OptStrBuf
{
   char   *StrPtr;                               // -> beginning of string.  May be NULL if       //
                                                 // string buffer does not currently store a      //
                                                 // string.                                       //
   int     UseCnt;                               // String buffer usage count.  Indicates the     //
                                                 // number of string objects that use the string  //
                                                 // buffer.  This count should always be greater  //
                                                 // than 0.  When the count goes to 0, the string //
                                                 // buffer will delete itself.                    //
   int     Len;                                  // Current string length.  The length may be 0.  //
                                                 // The length must not exceed the size of the    //
                                                 // buffer specified by the Siz member.           //
   int     Siz;                                  // Current buffer storage space size.  The size  //
                                                 // may be 0.  The size must be greater than or   //
                                                 // equal to the length specified by the Len      //
                                                 // member.                                       //
   friend  OptStr;
public:
   void Chk() const;
   OptStrBuf();
   OptStrBuf(const char *Str,int StrLen = -1,int StrSiz = -1);
   ~OptStrBuf();
   void Inc();
   void Dec();
   void ReSiz(int NewLen,int NewSiz);
private:
   OptStrBuf(const OptStrBuf &Src);
   OptStrBuf &operator =(const OptStrBuf &Rht);
};
////
class OptStr
{
   int        Len;                               // Current string length.  The length may be 0.  //
                                                 // The sum of the length and offset must not     //
                                                 // exceed the string buffer's length or size.    //
   int        Off;                               // Offset to start of string in string buffer.   //
                                                 // The sum of the length and offset must not     //
                                                 // exceed the string buffer's length or size.    //
   OptStrBuf *BufPtr;                            // -> string buffer storing the string.  This    //
                                                 // may be NULL if the string object stores no    //
                                                 // string.                                       //
public:
   void Chk() const;
   OptStr();
   OptStr(const OptStr &Src);
   OptStr(const char *Str,int StrLen = -1);
   ~OptStr();
   int length(void) const;
   OptStr &operator = (const OptStr &Src);
   OptStr &operator += (const OptStr &Str);
   OptStr &operator += (char Chr);
   const char &operator [] (int Ndx) const;
   char &operator [] (int Ndx);
   char *GetPtr(int Ndx);
   const char *GetStrPtr(int Ndx) const;
   const char *data() const;
   char *GetStrPtr(int Ndx);
   void MakSubStr(int SubOff,int SubLen);
   OptStr operator () (int SubOff,int SubLen) const;
   OptStr substr (int SubOff,int SubLen) const;
   void RTrm();
   OptStr RTrim() const;
   void LTrm();
   OptStr LTrim() const;
   void Trm();
   OptStr Trim() const;
   void PreWrt(int NewSiz,bool CpyFlg);
};
////
inline int MaxInt(int i,int j)
{
   return (i > j)?i:j;
}
////
inline int MinInt(int i,int j)
{
   return (i < j)?i:j;
}
////
void ErrDspMsg(const char *Msg)
{
//   cout << endl << Msg << endl;
   throw Msg;
}
////
#ifndef NDEBUG
#define StrBufChk(Buf) ((Buf).Chk())
#define RefStrChk(Str) ((Str).Chk())
#define OptStrChk(Str) ((Str).Chk())
#define New            new
#define Del            delete
#else
#define StrBufChk(Buf)
#define RefStrChk(Str)
#define OptStrChk(Str)
#define New            new
#define Del            delete
#endif
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  Check a String Buffer object for Errors                                                //
//      This procedure can be used to check a String Buffer object for errors.  This can be used  //
// to detects most occurrences of corrupted or invalid String Buffer objects.                     //
//                                                                                                //
//      The debug version of the library procedures usually call this procedure to check all the  //
// String Buffer objects passed to them and to check any String Buffer objects that are changed   //
// or are to returned to the caller.  These checks may cause the debug version of the library to  //
// run substantially slower than the release version.                                             //
void
RefStrBuf::Chk() const
{

   if (Siz && !StrPtr)                           // If string buffer has a size but no string,    //
      ErrDspMsg("String buffer has size and no storage space."); // Display an error message.     //
   if (Len > Siz)                                // If string length is greater than buffer size, //
      ErrDspMsg("String buffer length exceeds size."); // Display an error message.               //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  String Buffer Object Default Constructor                                               //
//      This procedure initializes a String Buffer object that does not store a string.  The      //
// procedure assumes that some string ojbect will be using the buffer so it initializes the usage //
// count to one.                                                                                  //
RefStrBuf::RefStrBuf() :
   StrPtr(NULL),                                 // Indicate there is no storage space.           //
   UseCnt(1),                                    // Initialize usage count to 1.                  //
   Len(0),                                       // Indicate nothing is stored.                   //
   Siz(0)                                        // Indicate there is no storage space.           //
{
   StrBufChk(*this);                             // Check this string buffer.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  Construct a String Buffer Object From a Constant Character Array                       //
//      This procedure initializes a String Buffer object from a constant character array. The    //
// procedure assumes that some string ojbect will be using the buffer so it initializes the usage //
// count to one.                                                                                  //
//                                                                                                //
//      If a NUL terminated string is used to specify the initial value, then the size and length //
// can be calculated from the length of the string.                                               //
RefStrBuf::RefStrBuf(const char *Str,            // -> string to initialize from.  This pointer   //
                                                 // may be NULL if the length and size both are 0 //
                                                 // or -1.                                        //
                    int          StrLen,         // Default = -1.  String length.  If a NUL       //
                                                 // terminated string is specified, then -1 may   //
                                                 // be specified to use the length of the string. //
                    int          StrSiz)         // Default = -1.  String Size.  If a NUL         //
                                                 // terminated string is specified, then -1 may   //
                                                 // be specified to use the length of the string. //
{
   if (StrLen == -1)                             // If no length is specified, then               //
      Len = Str?strlen(Str):0;                   // Get the length from the C-string.             //
   else                                          // Otherwise if the length is specified, then    //
      Len = StrLen;                              // Set the length to the specified length.       //
   if (StrSiz == -1)                             // If no size is specified, then                 //
      Siz = Len;                                 // Set the size to the Length.                   //
   else                                          // Otherwise if the size is specified, then      //
      Siz = StrSiz;                              // Set the size to specified size.               //
   StrPtr = New char[Siz];                       // Allocate the string storage space.            //
   UseCnt = 1;                                   // Initialize usage count to 1.                  //
   memcpy(StrPtr,Str,Len);                       // Copy the string to allocated storage space.   //

   StrBufChk(*this);                             // Check this string buffer.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  String Buffer object Destructor                                                        //
//      This procedure destroys a String Buffer object .  The procedure frees the dynamic         //
// memory that is used to store the string.                                                       //
RefStrBuf::~RefStrBuf()
{
   Del [] StrPtr;                                // Release the allocated storage space.          //

   StrBufChk(*this);                             // Check this string buffer.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  Increment String Buffer Object's Usage Count                                       //
//      This procedure increments a String Buffer object's usage count, indicating that the   //
// buffer is being used by an additional string ojbect.                                          //
void
RefStrBuf::Inc()
{
   StrBufChk(*this);                             // Check this string buffer.                     //

   ++UseCnt;                                     // Increment the usage count.                    //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  Decrement String Buffer Object's Usage Count                                           //
//      This procedure decrements a String Buffer object's usage count, indicating that the       //
// buffer is being used by one less string ojbect.  If the usage count goes to 0, indicating the  //
// string buffer is not being used by any String Objects, the procedure deallocates the string buffer. //
void
RefStrBuf::Dec()
{
   StrBufChk(*this);                             // Check this string buffer.                     //

   --UseCnt;                                     // Decrement the usage count.                    //
   if (UseCnt == 0)                              // If buffer is not being used by any string objects, //
      Del this;                                  // Destroy the string buffer.                    //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  Resize the String Buffer                                                               //
//      This procedure resizes a string buffer to hold a different length string.  The resized    //
// buffer will be at least as large as the requested size, but may be larger.                     //
//                                                                                                //
//      The procedure uses the new string length specified in the "NewLen" parameter to determine //
// how many characters should be copied from the old string buffer to the new one.  (The number   //
// of characters copied is actually the lesser of the new length and the old length.)  The        //
// procedure operates more quickly when this number is smaller or even 0, so resize operations    //
// should only specify the number of characters that need to be copied from the old buffer to the //
// new, in some cases this will be less than the actual string length.                            //
//                                                                                                //
//      The procedure tries to optimize the resize process so that needless resizes aren't        //
// performed.  For example, there are many string processing algorithms that tend to repeatidly   //
// add a single character onto the end of the string.  This tends to yield a lot of resizes as    //
// the string buffer's length is increased to make room for an additional character.  This        //
// procedure makes these cases more efficient by enlarging the buffer by more than the requested  //
// size on some of the resizes and then by not enlarging at all on the others.                    //
void
RefStrBuf::ReSiz(int NewLen,                     // New length of string stored in the string     //
                                                 // buffer.  This length must not exceed the      //
                                                 // requested buffer size.                        //
                 int NewSiz)                     // Requested size of string buffer.  The resized //
                                                 // string buffer may be bigger than requested    //
                                                 // size, but never smaller.                      //
{
   StrBufChk(*this);                             // Check this string buffer.                     //

   const int SizChg = NewSiz - Siz;              // Requested change in string size.              //
   const int AloUnt = NewSiz >> 1;               // Allocation unit, unit of change in string     //
                                                 // storage space.                                //
   if (SizChg > 0)                               // If buffer size is being increaced, then       //
   {
      if (SizChg < AloUnt)                       // If less than an allocation is requested, then //
         NewSiz = NewSiz + AloUnt;               // Increase size by an allocation unit.          //
   }
   else if (SizChg > -AloUnt || SizChg > -64)    // If string buffer is being shortened by more   //
                                                 // than an allocation unit or 64 bytes, then     //
         NewSiz = Siz;                           // Set the size to the specified new size.       //
   Len = MinInt(MinInt(Len,NewLen),NewSiz);      // Set the length to the smallest value of the   //
                                                 // length, new length or new size.               //
   if (NewSiz != Siz)                            // If the buffers size is not changing, then     //
   {
      char *NewStrPtr = New char[NewSiz];        // Allocate memory for new string buffer.        //

      memcpy(NewStrPtr,StrPtr,Len);              // Copy the string to the new storage space.     //
      Siz = NewSiz;                              // Set the buffer size to specified new size.    //
      Del [] StrPtr;                             // Release the old storage space.                //
      StrPtr = NewStrPtr;                        // Set pointer to point to the new buffer.       //
   }
}
////
//                                                                                                //
//                           String Object Procedures                                             //
//                                                                                                //
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMSC                                                                                 //
// TITLE:  Check a String Object for Errors                                                       //
//      This procedure can be used to check a string object for errors.  This can be used to      //
// detect most occurrences of corrupted or invalid string objects.                                //
//                                                                                                //
//      The debug version of the library procedures usually call this procedure to check all the  //
// string objects passed to them and to check any string objects that are changed or are to       //
// returned to the caller. These checks may cause the debug version of the library to run         //
// substantially slower than the release version.                                                 //
void
RefStr::Chk() const                              // Line number of caller, used if an error       //
                                                 // message is displayed.                         //
{
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTCST                                                                                 //
// TITLE:  String Object Default Constructor                                                      //
//      This procedure initializes a string object that does not store a string.  This constuctor //
// is very fast and efficient.  It creates a string object that uses a special representation of  //
// an empty string, however, the string object can be used in every way like a regular string     //
// object.                                                                                        //
RefStr::RefStr() :
   BufPtr(NULL)                                  // Indicate there is no string storage buffer.   //
{
   RefStrChk(*this);                             // Check this string object.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTCST                                                                                 //
// TITLE:  String Object Copy Constructor                                                         //
//      This procedure initializes a string object to store a copy of a string stored in another  //
// string object.  This constructor is very fast and efficient.                                   //
RefStr::RefStr(const RefStr &Src) :              // Reference to string object to initialize      //
                                                 // from.                                         //
   BufPtr(Src.BufPtr)                            // Set -> source string buffer.                  //
{
   if (BufPtr)                                   // If there is a string buffer, then             //
      BufPtr->Inc();                             // Increment the string buffer's usage count.    //
   RefStrChk(*this);                             // Check this string object.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTCST                                                                                 //
// TITLE:  Construct a String Object From a Constant Character Array                              //
//      This procedure initializes a string object from a constant character array.               //
//                                                                                                //
//      If a NUL terminated string is used to specify the initial value, then the string length   //
// can be calculated from the length of the initial string.                                       //
RefStr::RefStr(const char *Str,                  // -> string to initialize from.  This pointer   //
                                                 // may not be NULL.                              //
               int         StrLen) :             // Default = -1.  String length.  If a NUL       //
                                                 // terminated string is specified in the "Str"   //
                                                 // parameter, then -1 may be specified to        //
                                                 // indicate that the the length of the NUL       //
                                                 // terminated string should be used.             //
   BufPtr(New RefStrBuf(Str,StrLen))             // Allocate the memory for string buffer.        //
{
   RefStrChk(*this);                             // Check this string object.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMSC                                                                                 //
// TITLE:  String Object Destructor                                                               //
//      This procedure destroys a string object.                                                  //
RefStr::~RefStr()
{
   RefStrChk(*this);                             // Check this string object.                     //
   if (BufPtr)                                   // If there is a string buffer, then             //
      BufPtr->Dec();                             // Decrement the string buffer's usage count.    //
}
////
// MBRPRC: PUBLIC INLINE                                                                          //
// TYPE:   QSTATT                                                                                 //
// TITLE:  Get String Object's Length                                                             //
//      This procedure returns the current length of the specified string object.                 //
int                                              // String's current length.                      //
RefStr::length(void) const
{
   RefStrChk(*this);                             // Check this string.                            //
   return BufPtr?BufPtr->Len:0;                  // Return current length.                        //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMAN                                                                                 //
// TITLE:  Assignment Operator (=) from a String Object                                           //
//      This procedure assigns the value of the string object on the right side of the operator   //
// to the string object on the left size of the operator.  This operation is very fast and        //
// efficient.                                                                                     //
RefStr &                                         // Reference to destination string object with   //
                                                 // new value.                                    //
RefStr::operator = (const RefStr &Src)           // Source string object.                         //
{
   RefStrChk(Src);                               // Check the source string object.               //
   RefStrChk(*this);                             // Check this string object.                     //

   if (this != &Src)                             // If this string object is not the source       //
                                                 // string object,                                //
   {
      if (BufPtr)                                // If this string has a string buffer, then      //
         BufPtr->Dec();                          // Decrement the string buffer's usage count.    //
      BufPtr = Src.BufPtr;                       // Get the source string's string buffer.        //
      if (BufPtr)                                // If there is a string buffer, then             //
         BufPtr->Inc();                          // Increment the string buffer's usage count.    //

      RefStrChk(*this);                          // Check this string object.                          //
   }
   return *this;                                 // Return destination string.                    //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMAN                                                                                 //
// TITLE:  Concatenation Assignment Operator (+=) for Two String Objects                          //
//      This procedure concatenates the string object on the right side of the operand onto the   //
// string object on the left.  The result is returned in the left string object.  The right       //
// string object is not changed by this procedure.  In the worst case, the duration of this       //
// operation is approximately proportional to the total length of the two source strings.         //
// However, in some cases the duration will depend only on the length of the string on the right. //
// Whenever possible, use this operation rather than the less efficient concatenation operator    //
// (+).                                                                                           //
RefStr &                                         // A reference to the left string object         //
                                                 // containing the concatenated string.           //
RefStr::operator += (const RefStr &Str)          // The string object to concatenate on.          //
{
   RefStrChk(*this);                             // Check left operand string.                    //
   RefStrChk(Str);                               // Check right operand string.                   //

   const int SrcLen = Str.length();

   PreWrt(length() + SrcLen,true);               // Resize the string.                            //
   if (SrcLen)                                   // If there is a left string.                    //
      memcpy(BufPtr->StrPtr+BufPtr->Len,         // Copy source string to destination buffer.     //
             Str.BufPtr->StrPtr,SrcLen);
   if (BufPtr)                                   // If there is a destination string buffer, then //
      BufPtr->Len+=SrcLen;                       // Calculate new string length.                  //
   return *this;                                 // Return destination string.                    //
}
////
// TITLE:  Concatenation Assignment Operator (+=) from a Character                                //
//      This procedure concatenates the character on the right side of the operand onto the       //
// string object on the left.  The result is returned in the left string object. Whenever         //
// possible, use this operation rather than the less efficient concatenation operator (+).        //
RefStr &                                         // A reference to the left string object         //
                                                 // containing the concatenated string.           //
RefStr::operator += (char Chr)                   // The character to concatenate on.              //
{
   RefStrChk(*this);                             // Check left operand string.                    //

   PreWrt(length() + 1,true);                    // Resize the string.                            //
   BufPtr->StrPtr[BufPtr->Len] = Chr;
   ++BufPtr->Len;                                // Calculate new string length.                  //
   RefStrChk(*this);                             // Check left operand string.                    //
   return *this;                                 // Return destination string.                    //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMAN                                                                                 //
// TITLE:  Index Operator ([]) for a Constant String Object                                       //
//      This procedure returns a reference to the constant character at the specified index in    //
// the string.  The specified index must be valid, that is, it must be between 0 and Len-1.  The  //
// character reference returned can be considered valid until the string is changed by any means. //
const char &                                     // Reference to constant character stored at the //
                                                 // specified offset in the string.               //
RefStr::operator [] (int Ndx) const              // The zero-relative index of the character to   //
                                                 // be returned.  If this index is invalid, the   //
                                                 // behavior is undefined.                        //
{
   RefStrChk(*this);                             // Check left operand.                           //

   return BufPtr->StrPtr[Ndx];                   // Return -> specified character.                //
}
////
// TITLE:  Index Operator ([]) for a String Object                                                //
//      This procedure returns a reference to the character at the specified index in the string. //
// The specified index must be valid, that is, it must be between 0 and Len-1.  The character     //
// reference returned may be changed to change the specified character in the string.  The        //
// character reference returned can be considered valid until the string is changed by any means  //
// other than changing the character referred to.                                                 //
char &                                           // Reference to character stored at the          //
                                                 // specified offset in the string.               //
RefStr::operator [] (int Ndx)                    // The zero-relative index of the character to   //
                                                 // be returned.  If this index is invalid, the   //
                                                 // behavior is undefined.                        //
{
   RefStrChk(*this);                             // Check left operand.                           //

   PreWrt(length(),true);                        // Get private copy of the string.               //
   return BufPtr->StrPtr[Ndx];                   // Return -> specified character.                //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMAN                                                                                 //
// TITLE:  Get Pointer to a Character in a String Object                                               //
//      This procedure returns a pointer to the character at the specified offset in the string.  //
// If the specified offset is not within the string (is not between 0 and Len-1), a NULL pointer  //
// is returned.  The pointer returned may be used to test or modify other characters in the       //
// string so long as it does no go past the ends of the buffer that stores the string.  The       //
// pointer returned can be considered valid until the string is changed by any means other than   //
// through this pointer.                                                                          //
char *                                           // -> character stored at the specified offset   //
                                                 // in the string.  NULL if specified offset is   //
                                                 // not within the string.                        //
RefStr::GetPtr(int Ndx)                          // The zero-relative index of the character to   //
                                                 // be returned.                                  //
{
   RefStrChk(*this);                             // Check this string.                            //

   if (Ndx < 0 || length() <= Ndx )              // If index is not within the string, then       //
      return NULL;                               // Return a NULL pointer.                        //
   PreWrt(length(),true);                        // Get private copy of the string.               //
   return BufPtr->StrPtr+Ndx;                    // Return -> specified character.                //
}
////
// TITLE:  Get Pointer to a Character in a Constant String Object                                      //
//      This procedure returns a pointer to the constant character at the specified offset in the //
// string.  If the specified offset is not within the string (is not between 0 and Len-1), a NULL //
// pointer is returned.  The pointer returned may be used to test other characters in the string  //
// so long as it does no go past the ends of the buffer that stores the string.  The pointer      //
// returned can be considered valid until the string is changed in any way.                       //
const char *                                     // -> constant character stored at the specified //
                                                 // offset in the string.  NULL if specified      //
                                                 // offset is not within the string.              //
RefStr::GetStrPtr(int Ndx) const                // The zero-relative index of the character to   //
                                                 // be returned.                                  //
{
   RefStrChk(*this);                             // Check this string.                            //

   if (Ndx < 0 || length() <= Ndx)               // If index is not within the string, then       //
      return NULL;                               // Return a NULL pointer.                        //
   return BufPtr->StrPtr+Ndx;                    // Return -> specified character.                //
}
////
// TITLE:  Get Pointer to Start of String Text                                                    //
//      This procedure returns a pointer to the start of the string text.                         //
const char *                                     // -> start of the string.                       //
RefStr::data() const
{
   if (BufPtr)                                   // If index is not within the string, then       //
      return BufPtr->StrPtr;                     // Return -> string data.                        //
   return NULL;                                  // Return a NULL pointer.                        //
}
////
// MBRPRC: PUBLIC INLINE                                                                          //
// TYPE:   QSTMAN                                                                                 //
// TITLE:  Get Pointer to a Character in a Constant String Object                                      //
//      This procedure returns a pointer to the constant character at the specified offset in the //
// string.  If the specified offset is not within the string (is not between 0 and Len-1), a NULL //
// pointer is returned.  The pointer returned may be used to test or modify other characters in   //
// the string so long as it does no go past the ends of the buffer that stores the string.  The       //
// pointer returned can be considered valid until the string is changed by any means other than   //
// through this pointer.
char *                                           // -> constant character stored at the specified //
                                                 // offset in the string.  NULL if specified      //
                                                 // offset is not within the string.              //
RefStr::GetStrPtr(int Ndx)                        // The zero-relative index of the character to   //
                                                 // be returned.                                  //
{
   RefStrChk(*this);                             // Check this string.                        //

   if (Ndx < 0 || length() <= Ndx )              // If index is not within the string, then       //
      return NULL;                               // Return a NULL pointer.                        //

   return BufPtr->StrPtr+Ndx;                    // Return -> specified character.                //
}
////
// TITLE:  Make Sub-String of a String Object                                                     //
//      This procedure changes a string object to a sub-string of its previous value.  The new    //
// string starts at the specified position and extends to the specified length.  If part or all   //
// of the specified sub-string extends beyond the ends of the source string, only the portion     //
// that is within the source string (if any) is returned.  The sub-string operator (parenthesis)  //
// is similar except it returns a new string that is a sub-string of the source string.           //
//                                                                                                //
//      The substring procedure is slightly more efficient than the sub-string operator and       //
// should be prefered in cases where either could be used.                                        //
void
RefStr::MakSubStr(int SubOff,                    // The zero-relative offset where the sub-string //
                                                 // should start.  If the offset indicates a      //
                                                 // position past the end of the string, a string //
                                                 // of zero length will be returned.  If the      //
                                                 // offset is less than zero, only the portion of //
                                                 // the string past position 0, if any, will be   //
                                                 // returned.                                     //
                  int SubLen)                    // The length of the sub-string to be returned.  //
                                                 // If part or all of the sub-string extends      //
                                                 // beyond the ends of the source string, then    //
                                                 // the string returned will be shorter.          //
{
   RefStrChk(*this);                             // Check string being sub-stringed.              //

   if (!BufPtr)                                  // If there is no string, then                   //
      return;                                    // Return without doing anyting.                 //
   if (SubOff < 0)                               // If offset is before beginning of string.      //
   {
      SubLen += SubOff;                          // Adjust length to start from beginning.        //
      SubOff = 0;                                // Start at beginning.                           //
   }
   SubLen = MaxInt(0,MinInt(SubLen,length()-SubOff)); // Calculate actual sub-string length.      //
   if (SubLen)                                   // If there is a sub-string, then                //
   {
      if (BufPtr->UseCnt > 1)                    // If this string buffer is shared, then         //
      {
         RefStrBuf *NewBufPtr = new RefStrBuf(BufPtr->StrPtr+SubOff,SubLen,SubLen);
         BufPtr->Dec();                          // Decrement old string buffer's usage count.    //
         BufPtr = NewBufPtr;                     // Switch to the new string buffer.              //
      }
      else                                       // Otherwise, if string buffer is not shared,    //
      {
         if (SubOff)                             // If there is an offset,                        //
            memmove(BufPtr->StrPtr,BufPtr->StrPtr+SubOff,SubLen); // Copy string to start of buffer. //
         BufPtr->Len = SubLen;                   // Reset the string length.                      //
      }
   }
   else                                          // Otherise, if the length is 0, then            //
   {
      BufPtr->Dec();                             // Decrement the string buffer's usage count.    //
      BufPtr = NULL;                             // Indicate there is no string buffer.
   }
   RefStrChk(*this);                             // Check sub-string formed.                      //
}
////
// TITLE:  Sub-String Operator (()) for a String Object                                           //
//      This procedure returns a string object that is the sub-string of the specified string.    //
// The sub-string returned starts at the specified position and extends to the specified length.  //
// If part or all of the specified sub-string extends beyond the ends of the source string, only  //
// the portion that is within the source string (if any) is returned.  The source string is not   //
// changed by this operation.                                                                     //
//                                                                                                //
//      The substring procedure is slightly more efficient than the sub-string operator and       //
// should be prefered in cases where either could be used.                                        //
RefStr                                           // The sub-string that starts at the specified   //
                                                 // position and extends for the specified        //
                                                 // length.                                       //
RefStr::operator () (int SubOff,                 // The zero-relative offset where the sub-string //
                                                 // should start.  If the offset indicates a      //
                                                 // position past the end of the string, a string //
                                                 // of zero length will be returned.  If the      //
                                                 // offset is less than zero, only the portion of //
                                                 // the string past position 0, if any, will be   //
                                                 // returned.                                     //
                     int SubLen) const           // The length of the sub-string to be returned.  //
                                                 // If part or all of the sub-string extends      //
                                                 // beyond the ends of the source string, then    //
                                                 // the string returned will be shorter.          //
{
   RefStrChk(*this);                             // Check string being sub-stringed.              //

   RefStr Dst(*this);                            // Construct string object to be returned.       //

   Dst.MakSubStr(SubOff,SubLen);                 // Sub-string the return value.                  //
   return Dst;                                   // Return destination string.                    //
}
////
// TITLE:  Sub-String Operator (()) for a String Object                                           //
//      This procedure returns a string object that is the sub-string of the specified string.    //
// The sub-string returned starts at the specified position and extends to the specified length.  //
// If part or all of the specified sub-string extends beyond the ends of the source string, only  //
// the portion that is within the source string (if any) is returned.  The source string is not   //
// changed by this operation.                                                                     //
//                                                                                                //
//      The substring procedure is slightly more efficient than the sub-string operator and       //
// should be prefered in cases where either could be used.                                        //
RefStr                                           // The sub-string that starts at the specified   //
                                                 // position and extends for the specified        //
                                                 // length.                                       //
RefStr::substr(int SubOff,                       // The zero-relative offset where the sub-string //
                                                 // should start.  If the offset indicates a      //
                                                 // position past the end of the string, a string //
                                                 // of zero length will be returned.  If the      //
                                                 // offset is less than zero, only the portion of //
                                                 // the string past position 0, if any, will be   //
                                                 // returned.                                     //
               int SubLen) const                 // The length of the sub-string to be returned.  //
                                                 // If part or all of the sub-string extends      //
                                                 // beyond the ends of the source string, then    //
                                                 // the string returned will be shorter.          //
{
   RefStrChk(*this);                             // Check string being sub-stringed.              //

   RefStr Dst(*this);                            // Construct string object to be returned.       //

   Dst.MakSubStr(SubOff,SubLen);                 // Sub-string the return value.                  //
   return Dst;                                   // Return destination string.                    //
}
////
// TITLE:  Remove Trailing Spaces From a String                                                   //
//      This procedure returns removes the trailing spaces from a string object.  This procedure  //
// is similar to the RTrim() procedure, except that this procedure changes the specified string   //
// rather than creating a new string.  In cases where a copy of the untrimmed string is not       //
// needed, use RTrm() over RTrim() as it is more efficient.                                       //
void
RefStr::RTrm()
{
   RefStrChk(*this);                             // Check this string.                            //

   int         Len = length();                   // String length.                                //
   const char *p   = GetStrPtr(Len - 1);         // Pointer to end of trimmed string.             //

   while (Len > 0 && *p == ' ')                  // While current character is a space,           //
   {
      --Len;                                     // Reduce length by one.                         //
      --p;                                       // Point to previous character.                  //
   }
   MakSubStr(0,Len);                             // Form a substring with portion to be retained. //
   RefStrChk(*this);                             // Check this string.                            //
}
////
// TITLE:  Remove Trailing Spaces From a String                                                   //
//      This procedure returns a string that is the same as "this" string except that the         //
// trailing spaces have been removed.  The source string is not affected by this procedure.  The  //
// RTrm() procedure is similar, except it trims the specified source string, rather than          //
// returning a new trimmed string.  In cases where a copy of the untrimmed string is not needed,  //
// use RTrm() over RTrim() as it is more efficient.                                               //
RefStr                                           // The trimmed string.                           //
RefStr::RTrim() const
{
   RefStrChk(*this);                             // Check this string.                            //

   RefStr RetVal(*this);                         // Construct the new string object.              //

   RetVal.RTrm();                                // Trim string on the right.                     //

   RefStrChk(RetVal);                            // Check the new string object.                       //

   return RetVal;                                // return the string object.                          //
}
////
// TITLE:  Remove Leading Spaces From a String                                                    //
//      This procedure returns removes the leading spaces from a string object.  This procedure   //
// is similar to the LTrim() procedure, except that this procedure changes the specified string   //
// rather than creating a new string.  In cases where a copy of the untrimmed string is not       //
// needed, use LTrm() over LTrim() as it is more efficient.                                       //
void
RefStr::LTrm()
{
   RefStrChk(*this);                             // Check this string.                            //

   int         Len = length();                   // String length.                                //
   int         Off = 0;                          // Offset to portion to retain.                  //
   const char *p   = GetStrPtr(0);               // Pointer to start of trimmed string.           //

   while (Len > 0 && *p == ' ')                  // While current character is a space,           //
   {
      --Len;                                     // Reduce length by one.                         //
      ++p;                                       // Point to next character.                      //
      ++Off;                                     // change offset by one.                         //
   }
   MakSubStr(Off,Len);                           // Form a substring with portion to be retained. //
   RefStrChk(*this);                             // Check this string.                            //
}
////
// TITLE:  Remove Leading Spaces From a String                                                    //
//      This procedure returns a string that is the same as "this" string except that the leading //
// spaces have been removed.  The source string is not affected by this procedure.  The LTrm()    //
// procedure is similar, except it trims the specified source string, rather than returning a new //
// trimmed string.  In cases where a copy of the untrimmed string is not needed, use LTrm() over  //
// LTrim() as it is more efficient.                                                               //
RefStr                                           // The trimmed string.                           //
RefStr::LTrim() const
{
   RefStrChk(*this);                             // Check this string.                            //

   RefStr RetVal(*this);                         // Construct the this string object.             //

   RetVal.LTrm();                                // Trim the string on the left.                  //

   RefStrChk(RetVal);                            // Check string object formed.                        //

   return RetVal;                                // return the string object.                          //
}
////
// TITLE:  Remove Leading and Trailing Spaces From a String                                       //
//      This procedure returns removes the leading and trailing spaces from a string object.      //
// This procedure is similar to the Trim() procedure, except that this procedure changes the      //
// specified string rather than creating a new string.  In cases where a copy of the untrimmed    //
// string is not needed, use Trm() over Trim() as it is more efficient.                           //
void
RefStr::Trm()
{
   RTrm();                                       // Trim string on the right.                     //
   LTrm();                                       // Trim string on the left.                      //
}
////
// TITLE:  Remove Leading and Trailing Spaces From a String                                       //
//      This procedure returns a string that is the same as "this" string except that the leading //
// and trailing spaces have been removed.  The source string is not affected by this procedure.   //
// The Trm() procedure is similar, except it trims the specified source string, rather than       //
// returning a new trimmed string.  In cases where a copy of the untrimmed string is not needed,  //
// use Trm() over Trim() as it is more efficient.                                                 //
RefStr                                           // The trimmed string.                           //
RefStr::Trim() const
{
   return RTrim().LTrim();                       // Return string trimmed on the right and left.  //
}
////
// TITLE:  Prepare to Write String                                                                //
//      This procedure prepares a string to be changed (written to).  The procedure insures that  //
// the string uses a string buffer that is not shared with other strings.  In addition, it        //
// resizes the string buffer to at least the requested size (if necessary).  In addition, it      //
// copies the old string to the resized buffer if specified by the parameters.                    //
void
RefStr::PreWrt(int  NewSiz,                      // Requested buffer size.  The buffer may be     //
                                                 // made bigger than requested size.              //
               bool CpyFlg)                      // This flag specifies whether or not the old    //
                                                 // contents of the string buffer should be       //
                                                 // copied to the new, resized buffer.  Specify   //
                                                 // true to copy the string if the previous value //
                                                 // will still be needed.  Specify false if the   //
                                                 // old value will not be needed.                 //
{
   RefStrChk(*this);                             // Check this string object.                     //

   const int Len = CpyFlg?length():0;            // The length to be copied.                      //

   if (BufPtr && BufPtr->UseCnt == 1)            // If string buffer is not used by other strngs, //
   {
      BufPtr->ReSiz(Len,NewSiz);                 // Resize the string buffer.                     //
   }
   else                                          // Otherwise, if buffer is used by other         //
   {                                             // strings, or there is no buffer, then          //
      if (BufPtr)                                // If there a buffer, then                       //
      {
         BufPtr->Dec();                          // Decrement the usage count.                    //
         BufPtr = New RefStrBuf(BufPtr->StrPtr,  // Create a new string buffer with old value.    //
                                Len,NewSiz);
      }
      else if (NewSiz)                           // Otherwise, if there is no string buffer, then //
         BufPtr = New RefStrBuf(NULL,Len,NewSiz); // Create a new string buffer with no string.   //
   }
   RefStrChk(*this);                             // Check this string object.                          //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  Check a String Buffer object for Errors                                                //
//      This procedure can be used to check a String Buffer object for errors.  This can be used  //
// to detects most occurrences of corrupted or invalid String Buffer objects.                     //
//                                                                                                //
//      The debug version of the library procedures usually call this procedure to check all the  //
// String Buffer objects passed to them and to check any String Buffer objects that are changed   //
// or are to returned to the caller.  These checks may cause the debug version of the library to  //
// run substantially slower than the release version.                                             //
void
OptStrBuf::Chk() const
{

   if (Siz && !StrPtr)                           // If string buffer has a size but no string,    //
      ErrDspMsg("String buffer has size and no storage space."); // Display an error message.     //
   if (Len > Siz)                                // If string length is greater than buffer size, //
      ErrDspMsg("String buffer length exceeds size."); // Display an error message.               //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  String Buffer Object Default Constructor                                               //
//      This procedure initializes a String Buffer object that does not store a string.  The      //
// procedure assumes that some string ojbect will be using the buffer so it initializes the usage //
// count to one.                                                                                  //
OptStrBuf::OptStrBuf() :
   StrPtr(NULL),                                 // Indicate there is no storage space.           //
   UseCnt(1),                                    // Initialize usage count to 1.                  //
   Len(0),                                       // Indicate nothing is stored.                   //
   Siz(0)                                        // Indicate there is no storage space.           //
{
   StrBufChk(*this);                             // Check this string buffer.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  Construct a String Buffer Object From a Constant Character Array                       //
//      This procedure initializes a String Buffer object from a constant character array. The    //
// procedure assumes that some string ojbect will be using the buffer so it initializes the usage //
// count to one.                                                                                  //
//                                                                                                //
//      If a NUL terminated string is used to specify the initial value, then the size and length //
// can be calculated from the length of the string.                                               //
OptStrBuf::OptStrBuf(const char *Str,            // -> string to initialize from.  This pointer   //
                                                 // may be NULL if the length and size both are 0 //
                                                 // or -1.                                        //
                    int          StrLen,         // Default = -1.  String length.  If a NUL       //
                                                 // terminated string is specified, then -1 may   //
                                                 // be specified to use the length of the string. //
                    int          StrSiz)         // Default = -1.  String Size.  If a NUL         //
                                                 // terminated string is specified, then -1 may   //
                                                 // be specified to use the length of the string. //
{
   if (StrLen == -1)                             // If no length is specified, then               //
      Len = Str?strlen(Str):0;                   // Get the length from the C-string.             //
   else                                          // Otherwise if the length is specified, then    //
      Len = StrLen;                              // Set the length to the specified length.       //
   if (StrSiz == -1)                             // If no size is specified, then                 //
      Siz = Len;                                 // Set the size to the Length.                   //
   else                                          // Otherwise if the size is specified, then      //
      Siz = StrSiz;                              // Set the size to specified size.               //
   StrPtr = New char[Siz];                       // Allocate the string storage space.            //
   UseCnt = 1;                                   // Initialize usage count to 1.                  //
   memcpy(StrPtr,Str,Len);                       // Copy the string to allocated storage space.   //

   StrBufChk(*this);                             // Check this string buffer.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  String Buffer object Destructor                                                        //
//      This procedure destroys a String Buffer object .  The procedure frees the dynamic         //
// memory that is used to store the string.                                                       //
OptStrBuf::~OptStrBuf()
{
   Del [] StrPtr;                                // Release the allocated storage space.          //

   StrBufChk(*this);                             // Check this string buffer.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  Increment String Buffer Object's Usage Count                                       //
//      This procedure increments a String Buffer object's usage count, indicating that the   //
// buffer is being used by an additional string ojbect.                                          //
void
OptStrBuf::Inc()
{
   StrBufChk(*this);                             // Check this string buffer.                     //

   ++UseCnt;                                     // Increment the usage count.                    //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  Decrement String Buffer Object's Usage Count                                           //
//      This procedure decrements a String Buffer object's usage count, indicating that the       //
// buffer is being used by one less string ojbect.  If the usage count goes to 0, indicating the  //
// string buffer is not being used by any String Objects, the procedure deallocates the string buffer. //
void
OptStrBuf::Dec()
{
   StrBufChk(*this);                             // Check this string buffer.                     //

   --UseCnt;                                     // Decrement the usage count.                    //
   if (UseCnt == 0)                              // If buffer is not being used by any string objects, //
      Del this;                                  // Destroy the string buffer.                    //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   STRBUF                                                                                 //
// TITLE:  Resize the String Buffer                                                               //
//      This procedure resizes a string buffer to hold a different length string.  The resized    //
// buffer will be at least as large as the requested size, but may be larger.                     //
//                                                                                                //
//      The procedure uses the new string length specified in the "NewLen" parameter to determine //
// how many characters should be copied from the old string buffer to the new one.  (The number   //
// of characters copied is actually the lesser of the new length and the old length.)  The        //
// procedure operates more quickly when this number is smaller or even 0, so resize operations    //
// should only specify the number of characters that need to be copied from the old buffer to the //
// new, in some cases this will be less than the actual string length.                            //
//                                                                                                //
//      The procedure tries to optimize the resize process so that needless resizes aren't        //
// performed.  For example, there are many string processing algorithms that tend to repeatidly   //
// add a single character onto the end of the string.  This tends to yield a lot of resizes as    //
// the string buffer's length is increased to make room for an additional character.  This        //
// procedure makes these cases more efficient by enlarging the buffer by more than the requested  //
// size on some of the resizes and then by not enlarging at all on the others.                    //
void
OptStrBuf::ReSiz(int NewLen,                     // New length of string stored in the string     //
                                                 // buffer.  This length must not exceed the      //
                                                 // requested buffer size.                        //
                 int NewSiz)                     // Requested size of string buffer.  The resized //
                                                 // string buffer may be bigger than requested    //
                                                 // size, but never smaller.                      //
{
   StrBufChk(*this);                             // Check this string buffer.                     //

   const int SizChg = NewSiz - Siz;              // Requested change in string size.              //
   const int AloUnt = NewSiz >> 1;               // Allocation unit, unit of change in string     //
                                                 // storage space.                                //
   if (SizChg > 0)                               // If buffer size is being increaced, then       //
   {
      if (SizChg < AloUnt)                       // If less than an allocation is requested, then //
         NewSiz = NewSiz + AloUnt;               // Increase size by an allocation unit.          //
   }
   else if (SizChg > -AloUnt || SizChg > -64)    // If string buffer is being shortened by more   //
                                                 // than an allocation unit or 64 bytes, then     //
         NewSiz = Siz;                           // Set the size to the specified new size.       //
   Len = MinInt(MinInt(Len,NewLen),NewSiz);      // Set the length to the smallest value of the   //
                                                 // length, new length or new size.               //
   if (NewSiz != Siz)                            // If the buffers size is not changing, then     //
   {
      char *NewStrPtr = New char[NewSiz];        // Allocate memory for new string buffer.        //

      memcpy(NewStrPtr,StrPtr,Len);              // Copy the string to the new storage space.     //
      Siz = NewSiz;                              // Set the buffer size to specified new size.    //
      Del [] StrPtr;                             // Release the old storage space.                //
      StrPtr = NewStrPtr;                        // Set pointer to point to the new buffer.       //
   }
}
////
//                                                                                                //
//                           String Object Procedures                                             //
//                                                                                                //
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMSC                                                                                 //
// TITLE:  Check a String Object for Errors                                                       //
//      This procedure can be used to check a string object for errors.  This can be used to      //
// detect most occurrences of corrupted or invalid string objects.                                //
//                                                                                                //
//      The debug version of the library procedures usually call this procedure to check all the  //
// string objects passed to them and to check any string objects that are changed or are to       //
// returned to the caller. These checks may cause the debug version of the library to run         //
// substantially slower than the release version.                                                 //
void
OptStr::Chk() const                              // Line number of caller, used if an error       //
                                                 // message is displayed.                         //
{
   if (Len && !BufPtr)                           // If string object has length, but no buffer, then   //
      ErrDspMsg("String has size and no storage space."); // Display an error message.            //
   if (BufPtr && Len + Off > BufPtr->Len)        // If end of string is past buffer size, then    //
      ErrDspMsg("String length and offset exceeds string buffer length."); // Display an error message. //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTCST                                                                                 //
// TITLE:  String Object Default Constructor                                                      //
//      This procedure initializes a string object that does not store a string.  This constuctor //
// is very fast and efficient.  It creates a string object that uses a special representation of  //
// an empty string, however, the string object can be used in every way like a regular string     //
// object.                                                                                        //
OptStr::OptStr() :
   BufPtr(NULL),                                 // Indicate there is no string storage buffer.   //
   Off(0),                                       // Initialize offset to 0.                       //
   Len(0)                                        // Indicate nothing is stored.                   //
{
   OptStrChk(*this);                             // Check this string object.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTCST                                                                                 //
// TITLE:  String Object Copy Constructor                                                         //
//      This procedure initializes a string object to store a copy of a string stored in another  //
// string object.  This constructor is very fast and efficient.                                   //
OptStr::OptStr(const OptStr &Src) :              // Reference to string object to initialize      //
                                                 // from.                                         //
   BufPtr(Src.BufPtr),                           // Set -> source string buffer.                  //
   Len(Src.Len),                                 // Set the length to source string length.       //
   Off(Src.Off)                                  // Set the offset to source string offset.       //
{
   if (BufPtr)                                   // If there is a string buffer, then             //
      BufPtr->Inc();                             // Increment the string buffer's usage count.    //
   OptStrChk(*this);                             // Check this string object.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTCST                                                                                 //
// TITLE:  Construct a String Object From a Constant Character Array                              //
//      This procedure initializes a string object from a constant character array.               //
//                                                                                                //
//      If a NUL terminated string is used to specify the initial value, then the string length   //
// can be calculated from the length of the initial string.                                       //
OptStr::OptStr(const char *Str,                  // -> string to initialize from.  This pointer   //
                                                 // may not be NULL.                              //
               int         StrLen) :             // Default = -1.  String length.  If a NUL       //
                                                 // terminated string is specified in the "Str"   //
                                                 // parameter, then -1 may be specified to        //
                                                 // indicate that the the length of the NUL       //
                                                 // terminated string should be used.             //
   BufPtr(New OptStrBuf(Str,StrLen)),            // Allocate the memory for string buffer.        //
   Off(0)                                        // Set the offset 0.                             //
{
   Len = BufPtr->Len;                            // Set the length to string buffer's length.     //
   OptStrChk(*this);                             // Check this string object.                     //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMSC                                                                                 //
// TITLE:  String Object Destructor                                                               //
//      This procedure destroys a string object.                                                  //
OptStr::~OptStr()
{
   OptStrChk(*this);                             // Check this string object.                     //
   if (BufPtr)                                   // If there is a string buffer, then             //
      BufPtr->Dec();                             // Decrement the string buffer's usage count.    //
}
////
// MBRPRC: PUBLIC INLINE                                                                          //
// TYPE:   QSTATT                                                                                 //
// TITLE:  Get String Object's Length                                                             //
//      This procedure returns the current length of the specified string object.                 //
int                                              // String's current length.                      //
OptStr::length(void) const
{
   OptStrChk(*this);                             // Check this string.                            //
   return Len;                                   // Return current length.                        //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMAN                                                                                 //
// TITLE:  Assignment Operator (=) from a String Object                                           //
//      This procedure assigns the value of the string object on the right side of the operator   //
// to the string object on the left size of the operator.  This operation is very fast and        //
// efficient.                                                                                     //
OptStr &                                         // Reference to destination string object with   //
                                                 // new value.                                    //
OptStr::operator = (const OptStr &Src)           // Source string object.                         //
{
   OptStrChk(Src);                               // Check the source string object.               //
   OptStrChk(*this);                             // Check this string object.                     //

   if (this != &Src)                             // If this string object is not the source       //
                                                 // string object,                                //
   {
      if (BufPtr)                                // If this string has a string buffer, then      //
         BufPtr->Dec();                          // Decrement the string buffer's usage count.    //
      Len = Src.Len;                             // Get the source string's length.               //
      Off = Src.Off;                             // Get the source string's offset.               //
      BufPtr = Src.BufPtr;                       // Get the source string's string buffer.        //
      if (BufPtr)                                // If there is a string buffer, then             //
         BufPtr->Inc();                          // Increment the string buffer's usage count.    //

      OptStrChk(*this);                          // Check this string object.                          //
   }
   return *this;                                 // Return destination string.                    //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMAN                                                                                 //
// TITLE:  Concatenation Assignment Operator (+=) for Two String Objects                          //
//      This procedure concatenates the string object on the right side of the operand onto the   //
// string object on the left.  The result is returned in the left string object.  The right       //
// string object is not changed by this procedure.  In the worst case, the duration of this       //
// operation is approximately proportional to the total length of the two source strings.         //
// However, in some cases the duration will depend only on the length of the string on the right. //
// Whenever possible, use this operation rather than the less efficient concatenation operator    //
// (+).                                                                                           //
OptStr &                                         // A reference to the left string object         //
                                                 // containing the concatenated string.           //
OptStr::operator += (const OptStr &Str)          // The string object to concatenate on.          //
{
   OptStrChk(*this);                             // Check left operand string.                    //
   OptStrChk(Str);                               // Check right operand string.                   //

   PreWrt(Len + Str.Len,true);                   // Resize the string.                            //
   if (Str.Len)                                  // If there is a left string.                    //
      memcpy(BufPtr->StrPtr+Len,                 // Copy source string to destination buffer.     //
             Str.BufPtr->StrPtr+Str.Off,Str.Len);
   if (BufPtr)                                   // If there is a destination string buffer, then //
      BufPtr->Len+=Str.Len;                      // Calculate new string length.                  //
   Len += Str.Len;                               // Calculate new length.                         //
   return *this;                                 // Return destination string.                    //
}
////
// TITLE:  Concatenation Assignment Operator (+=) From a Character                                //
//      This procedure concatenates the character on the right side of the operand onto the       //
// string object on the left.  The result is returned in the left string object.  Whenever        //
// possible, use this operation rather than the less efficient concatenation operator (+).        //
OptStr &                                         // A reference to the left string object         //
                                                 // containing the concatenated string.           //
OptStr::operator += (char Chr)                   // The character to concatenate on.              //
{
   OptStrChk(*this);                             // Check left operand string.                    //

   PreWrt(Len + 1,true);                         // Resize the string.                            //
   BufPtr->StrPtr[Len] = Chr;
   ++BufPtr->Len;                                // Calculate new string length.                  //
   ++Len;                                        // Calculate new length.                         //
   return *this;                                 // Return destination string.                    //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMAN                                                                                 //
// TITLE:  Index Operator ([]) for a Constant String Object                                       //
//      This procedure returns a reference to the constant character at the specified index in    //
// the string.  The specified index must be valid, that is, it must be between 0 and Len-1.  The  //
// character reference returned can be considered valid until the string is changed by any means. //
const char &                                     // Reference to constant character stored at the //
                                                 // specified offset in the string.               //
OptStr::operator [] (int Ndx) const              // The zero-relative index of the character to   //
                                                 // be returned.  If this index is invalid, the   //
                                                 // behavior is undefined.                        //
{
   OptStrChk(*this);                             // Check left operand.                           //

   return BufPtr->StrPtr[Off+Ndx];               // Return -> specified character.                //
}
////
// TITLE:  Index Operator ([]) for a String Object                                                //
//      This procedure returns a reference to the character at the specified index in the string. //
// The specified index must be valid, that is, it must be between 0 and Len-1.  The character     //
// reference returned may be changed to change the specified character in the string.  The        //
// character reference returned can be considered valid until the string is changed by any means  //
// other than changing the character referred to.                                                 //
char &                                           // Reference to character stored at the          //
                                                 // specified offset in the string.               //
OptStr::operator [] (int Ndx)                    // The zero-relative index of the character to   //
                                                 // be returned.  If this index is invalid, the   //
                                                 // behavior is undefined.                        //
{
   OptStrChk(*this);                             // Check left operand.                           //

   PreWrt(Len,true);                             // Get private copy of the string.               //
   return BufPtr->StrPtr[Off+Ndx];               // Return -> specified character.                //
}
////
// MBRPRC: PUBLIC                                                                                 //
// TYPE:   QSTMAN                                                                                 //
// TITLE:  Get Pointer to a Character in a String Object                                               //
//      This procedure returns a pointer to the character at the specified offset in the string.  //
// If the specified offset is not within the string (is not between 0 and Len-1), a NULL pointer  //
// is returned.  The pointer returned may be used to test or modify other characters in the       //
// string so long as it does no go past the ends of the buffer that stores the string.  The       //
// pointer returned can be considered valid until the string is changed by any means other than   //
// through this pointer.                                                                          //
char *                                           // -> character stored at the specified offset   //
                                                 // in the string.  NULL if specified offset is   //
                                                 // not within the string.                        //
OptStr::GetPtr(int Ndx)                          // The zero-relative index of the character to   //
                                                 // be returned.                                  //
{
   OptStrChk(*this);                             // Check this string.                            //

   if (Ndx < 0 || Len <= Ndx )                   // If index is not within the string, then       //
      return NULL;                               // Return a NULL pointer.                        //
   PreWrt(Len,true);                             // Get private copy of the string.               //
   return BufPtr->StrPtr+Off+Ndx;                // Return -> specified character.                //
}
////
// TITLE:  Get Pointer to a Character in a Constant String Object                                      //
//      This procedure returns a pointer to the constant character at the specified offset in the //
// string.  If the specified offset is not within the string (is not between 0 and Len-1), a NULL //
// pointer is returned.  The pointer returned may be used to test other characters in the string  //
// so long as it does no go past the ends of the buffer that stores the string.  The pointer      //
// returned can be considered valid until the string is changed in any way.                       //
const char *                                     // -> constant character stored at the specified //
                                                 // offset in the string.  NULL if specified      //
                                                 // offset is not within the string.              //
OptStr::GetStrPtr(int Ndx) const                // The zero-relative index of the character to   //
                                                 // be returned.                                  //
{
   OptStrChk(*this);                             // Check this string.                            //

   if (Ndx < 0 || Len <= Ndx)                    // If index is not within the string, then       //
      return NULL;                               // Return a NULL pointer.                        //
   return BufPtr->StrPtr+Off+Ndx;                // Return -> specified character.                //
}
////
// TITLE:  Get Pointer to Start of String Text                                                    //
//      This procedure returns a pointer to the start of the string text.                         //
const char *                                     // -> start of the string.                       //
OptStr::data() const
{
   if (BufPtr)                                   // If index is not within the string, then       //
      return BufPtr->StrPtr+Off;                 // Return -> string data.                        //
   return NULL;                                  // Return a NULL pointer.                        //
}
////
// MBRPRC: PUBLIC INLINE                                                                          //
// TYPE:   QSTMAN                                                                                 //
// TITLE:  Get Pointer to a Character in a Constant String Object                                      //
//      This procedure returns a pointer to the constant character at the specified offset in the //
// string.  If the specified offset is not within the string (is not between 0 and Len-1), a NULL //
// pointer is returned.  The pointer returned may be used to test or modify other characters in   //
// the string so long as it does no go past the ends of the buffer that stores the string.  The       //
// pointer returned can be considered valid until the string is changed by any means other than   //
// through this pointer.
char *                                           // -> constant character stored at the specified //
                                                 // offset in the string.  NULL if specified      //
                                                 // offset is not within the string.              //
OptStr::GetStrPtr(int Ndx)                        // The zero-relative index of the character to   //
                                                 // be returned.                                  //
{
   OptStrChk(*this);                             // Check this string.                        //

   if (Ndx < 0 || Len <= Ndx )                   // If index is not within the string, then       //
      return NULL;                               // Return a NULL pointer.                        //

   return BufPtr->StrPtr+Off+Ndx;                // Return -> specified character.                //
}
////
// TITLE:  Make Sub-String of a String Object                                                     //
//      This procedure changes a string object to a sub-string of its previous value.  The new    //
// string starts at the specified position and extends to the specified length.  If part or all   //
// of the specified sub-string extends beyond the ends of the source string, only the portion     //
// that is within the source string (if any) is returned.  The sub-string operator (parenthesis)  //
// is similar except it returns a new string that is a sub-string of the source string.           //
//                                                                                                //
//      The substring procedure is slightly more efficient than the sub-string operator and       //
// should be prefered in cases where either could be used.                                        //
void
OptStr::MakSubStr(int SubOff,                    // The zero-relative offset where the sub-string //
                                                 // should start.  If the offset indicates a      //
                                                 // position past the end of the string, a string //
                                                 // of zero length will be returned.  If the      //
                                                 // offset is less than zero, only the portion of //
                                                 // the string past position 0, if any, will be   //
                                                 // returned.                                     //
                  int SubLen)                    // The length of the sub-string to be returned.  //
                                                 // If part or all of the sub-string extends      //
                                                 // beyond the ends of the source string, then    //
                                                 // the string returned will be shorter.          //
{
   OptStrChk(*this);                             // Check string being sub-stringed.              //

   if (SubOff < 0)                               // If offset is before beginning of string.      //
   {
      SubLen += SubOff;                          // Adjust length to start from beginning.        //
      SubOff = 0;                                // Start at beginning.                           //
   }
   SubLen = MaxInt(0,MinInt(SubLen,Len-SubOff)); // Calculate actual sub-string length.           //

   Len = SubLen;                                 // Set the sub-string's length.                  //
   if (Len)                                      // If the length is not 0, then                  //
      Off += SubOff;                             // Set the sub-string's starting offset.         //
   else                                          // Oherwise, if length is 0, then                //
      Off = 0;                                   // Zero the sub-string's starting offset.        //

   OptStrChk(*this);                             // Check sub-string formed.                      //
}
////
// TITLE:  Sub-String Operator (()) for a String Object                                           //
//      This procedure returns a string object that is the sub-string of the specified string.    //
// The sub-string returned starts at the specified position and extends to the specified length.  //
// If part or all of the specified sub-string extends beyond the ends of the source string, only  //
// the portion that is within the source string (if any) is returned.  The source string is not   //
// changed by this operation.                                                                     //
//                                                                                                //
//      The substring procedure is slightly more efficient than the sub-string operator and       //
// should be prefered in cases where either could be used.                                        //
OptStr                                           // The sub-string that starts at the specified   //
                                                 // position and extends for the specified        //
                                                 // length.                                       //
OptStr::operator () (int SubOff,                 // The zero-relative offset where the sub-string //
                                                 // should start.  If the offset indicates a      //
                                                 // position past the end of the string, a string //
                                                 // of zero length will be returned.  If the      //
                                                 // offset is less than zero, only the portion of //
                                                 // the string past position 0, if any, will be   //
                                                 // returned.                                     //
                     int SubLen) const           // The length of the sub-string to be returned.  //
                                                 // If part or all of the sub-string extends      //
                                                 // beyond the ends of the source string, then    //
                                                 // the string returned will be shorter.          //
{
   OptStrChk(*this);                             // Check string being sub-stringed.              //

   OptStr Dst(*this);                            // Construct string object to be returned.       //

   Dst.MakSubStr(SubOff,SubLen);                 // Sub-string the return value.                  //
   return Dst;                                   // Return destination string.                    //
}
////
// TITLE:  Sub-String Operator (()) for a String Object                                           //
//      This procedure returns a string object that is the sub-string of the specified string.    //
// The sub-string returned starts at the specified position and extends to the specified length.  //
// If part or all of the specified sub-string extends beyond the ends of the source string, only  //
// the portion that is within the source string (if any) is returned.  The source string is not   //
// changed by this operation.                                                                     //
//                                                                                                //
//      The substring procedure is slightly more efficient than the sub-string operator and       //
// should be prefered in cases where either could be used.                                        //
OptStr                                           // The sub-string that starts at the specified   //
                                                 // position and extends for the specified        //
                                                 // length.                                       //
OptStr::substr(int SubOff,                       // The zero-relative offset where the sub-string //
                                                 // should start.  If the offset indica
tes a      //
                                                 // position past the end of the string, a string //
                                                 // of zero length will be returned.  If the      //
                                                 // offset is less than zero, only the portion of //
                                                 // the string past position 0, if any, will be   //
                                                 // returned.                                     //
               int SubLen) const                 // The length of the sub-string to be returned.  //
                                                 // If part or all of the sub-string extends      //
                                                 // beyond the ends of the source string, then    //
                                                 // the string returned will be shorter.          //
{
   OptStrChk(*this);                             // Check string being sub-stringed.              //

   OptStr Dst(*this);                            // Construct string object to be returned.       //

   Dst.MakSubStr(SubOff,SubLen);                 // Sub-string the return value.                  //
   return Dst;                                   // Return destination string.                    //
}
////
// TITLE:  Remove Trailing Spaces From a String                                                   //
//      This procedure returns removes the trailing spaces from a string object.  This procedure  //
// is similar to the RTrim() procedure, except that this procedure changes the specified string   //
// rather than creating a new string.  In cases where a copy of the untrimmed string is not       //
// needed, use RTrm() over RTrim() as it is more efficient.                                       //
void
OptStr::RTrm()
{
   OptStrChk(*this);                             // Check this string.                            //

   const char *p = GetStrPtr(Len - 1);           // Pointer to end of trimmed string.             //

   while (Len > 0 && *p == ' ')                  // While current character is a space,           //
   {
      --Len;                                     // Reduce length by one.                         //
      --p;                                       // Point to previous character.                  //
   }

   OptStrChk(*this);                             // Check this string.                            //
}
////
// TITLE:  Remove Trailing Spaces From a String                                                   //
//      This procedure returns a string that is the same as "this" string except that the         //
// trailing spaces have been removed.  The source string is not affected by this procedure.  The  //
// RTrm() procedure is similar, except it trims the specified source string, rather than          //
// returning a new trimmed string.  In cases where a copy of the untrimmed string is not needed,  //
// use RTrm() over RTrim() as it is more efficient.                                               //
OptStr                                           // The trimmed string.                           //
OptStr::RTrim() const
{
   OptStrChk(*this);                             // Check this string.                            //

   OptStr RetVal(*this);                         // Construct the new string object.              //

   RetVal.RTrm();                                // Trim string on the right.                     //

   OptStrChk(RetVal);                            // Check the new string object.                       //

   return RetVal;                                // return the string object.                          //
}
////
// TITLE:  Remove Leading Spaces From a String                                                    //
//      This procedure returns removes the leading spaces from a string object.  This procedure   //
// is similar to the LTrim() procedure, except that this procedure changes the specified string   //
// rather than creating a new string.  In cases where a copy of the untrimmed string is not       //
// needed, use LTrm() over LTrim() as it is more efficient.                                       //
void
OptStr::LTrm()
{
   OptStrChk(*this);                             // Check this string.                            //

   const char *p = GetStrPtr(0);                 // Pointer to start of trimmed string.           //

   while (Len > 0 && *p == ' ')                  // While current character is a space,           //
   {
      --Len;                                     // Reduce length by one.                         //
      ++p;                                       // Point to next character.                      //
      ++Off;                                     // change offset by one.                         //
   }

   OptStrChk(*this);                             // Check this string.                            //
}
////
// TITLE:  Remove Leading Spaces From a String                                                    //
//      This procedure returns a string that is the same as "this" string except that the leading //
// spaces have been removed.  The source string is not affected by this procedure.  The LTrm()    //
// procedure is similar, except it trims the specified source string, rather than returning a new //
// trimmed string.  In cases where a copy of the untrimmed string is not needed, use LTrm() over  //
// LTrim() as it is more efficient.                                                               //
OptStr                                           // The trimmed string.                           //
OptStr::LTrim() const
{
   OptStrChk(*this);                             // Check this string.                            //

   OptStr RetVal(*this);                         // Construct the this string object.             //

   RetVal.LTrm();                                // Trim the string on the left.                  //

   OptStrChk(RetVal);                            // Check string object formed.                        //

   return RetVal;                                // return the string object.                          //
}
////
// TITLE:  Remove Leading and Trailing Spaces From a String                                       //
//      This procedure returns removes the leading and trailing spaces from a string object.      //
// This procedure is similar to the Trim() procedure, except that this procedure changes the      //
// specified string rather than creating a new string.  In cases where a copy of the untrimmed    //
// string is not needed, use Trm() over Trim() as it is more efficient.                           //
void
OptStr::Trm()
{
   RTrm();                                       // Trim string on the right.                     //
   LTrm();                                       // Trim string on the left.                      //
}
////
// TITLE:  Remove Leading and Trailing Spaces From a String                                       //
//      This procedure returns a string that is the same as "this" string except that the leading //
// and trailing spaces have been removed.  The source string is not affected by this procedure.   //
// The Trm() procedure is similar, except it trims the specified source string, rather than       //
// returning a new trimmed string.  In cases where a copy of the untrimmed string is not needed,  //
// use Trm() over Trim() as it is more efficient.                                                 //
OptStr                                           // The trimmed string.                           //
OptStr::Trim() const
{
   return RTrim().LTrim();                       // Return string trimmed on the right and left.  //
}
////
// TITLE:  Prepare to Write String                                                                //
//      This procedure prepares a string to be changed (written to).  The procedure insures that  //
// the string uses a string buffer that is not shared with other strings.  In addition, it        //
// resizes the string buffer to at least the requested size (if necessary).  In addition, it      //
// copies the old string to the resized buffer if specified by the parameters.                    //
void
OptStr::PreWrt(int  NewSiz,                      // Requested buffer size.  The buffer may be     //
                                                 // made bigger than requested size.              //
               bool CpyFlg)                      // This flag specifies whether or not the old    //
                                                 // contents of the string buffer should be       //
                                                 // copied to the new, resized buffer.  Specify   //
                                                 // true to copy the string if the previous value //
                                                 // will still be needed.  Specify false if the   //
                                                 // old value will not be needed.                 //
{
   OptStrChk(*this);                             // Check this string object.                          //

   if (!CpyFlg)                                  // If string is not being copied, then           //
      Len = 0;                                   // Set the string length to 0.                   //

   if (BufPtr && BufPtr->UseCnt == 1)            // If string buffer is not used by other strngs, //
   {
      if (Off)                                   // If the memory needs to be shifted forwards,   //
         memmove(BufPtr->StrPtr,BufPtr->StrPtr+Off,Len); // Copy string to start of buffer.       //
      BufPtr->ReSiz(Len,NewSiz);                 // Resize the string buffer.                     //
   }
   else                                          // Otherwise, if buffer is used by other         //
   {                                             // strings, or there is no buffer, then          //
      if (BufPtr)                                // If there a buffer, then                       //
      {
         BufPtr->Dec();                          // Decrement the usage count.                    //
         BufPtr = New OptStrBuf(BufPtr->StrPtr+Off, // Create a new string buffer with old value. //
                                Len,NewSiz);
      }
      else if (NewSiz)                           // Otherwise, if there is no string buffer, then //
         BufPtr = New OptStrBuf(NULL,Len,NewSiz); // Create a new string buffer with no string.   //
   }
   Off = 0;                                      // Set the offset to 0.                          //
   OptStrChk(*this);                             // Check this string object.                          //
}
////
OptStr operator + (const OptStr &Lft,            // The string to concatenate on to.              //
                   const OptStr &Rht)            // The string to be concatenated on.             //
{
   OptStr Dst(Lft);                              // Construct destination string.                 //

   Dst += Rht;                                   // Add source string to destination string.      //
   return Dst;                                   // Return destination string.                    //
}
////
RefStr operator + (const RefStr &Lft,            // The string to concatenate on to.              //
                   const RefStr &Rht)            // The string to be concatenated on.             //
{
   RefStr Dst(Lft);                              // Construct destination string.                 //

   Dst += Rht;                                   // Add source string to destination string.      //
   return Dst;                                   // Return destination string.                    //
}
////

const int RepCnt = 100000;
const int TstCnt = 5;

enum StrTyp
{
   OptStrTyp,
   RefStrTyp,
   STLStrTyp,
   MaxStrTyp,
};

const char *StrTypNamLst[MaxStrTyp] = {"Optimized:","Reference:","STL:"};
enum TstTyp
{
  RecTstTyp,
  RevTstTyp,
  UprTstTyp,
  ChrTstTyp,
  WrpTstTyp,
  MaxTstTyp
};
const char *TstTypNamLst[MaxTstTyp] = {"Recipe:","Reversed:","Upper:","Char Search:","Wrap"};

DWORD Results[MaxTstTyp][MaxStrTyp][TstCnt];
////
template <class T>
T GetNextWord(const T &Str,
              int     &Pos)
{
   const int StrLen  = Str.length();
   int       BgnPos  = Pos;

   while (BgnPos < StrLen && Str[BgnPos] == ' ')
      ++BgnPos;

   int EndPos = BgnPos;

   while (EndPos < StrLen && Str[EndPos] != ' ')
      ++EndPos;
   Pos = EndPos;
   if (Pos >= StrLen)
      Pos = -1;
   return Str.substr(BgnPos,EndPos-BgnPos);
};
////
template <class T>
void RunRecTst(const T &Text,int TstNum,StrTyp Typ)
{
   DWORD BgnTim,EndTim;

   BgnTim = GetTickCount();
   for (int i = 0; i < RepCnt; ++i)
   {
      int Pos = 0;

      while (Pos >=0)
         GetNextWord(Text,Pos);
   }
   EndTim = GetTickCount();
   Results[RecTstTyp][Typ][TstNum] = EndTim-BgnTim;
};
////
template <class T>
void RunRevTst(const T &Text,int TstNum,StrTyp Typ)
{
   DWORD BgnTim,EndTim;

   BgnTim = GetTickCount();

   for (int i = 0; i < RepCnt; ++i)
   {
      int Pos = 0;

      T RevStr;

      while (Pos >=0)
         RevStr = GetNextWord(Text,Pos) + RevStr;
   }

   EndTim = GetTickCount();
   Results[RevTstTyp][Typ][TstNum] = EndTim-BgnTim;
};
////
template <class T>
void RunUprTst(const T &Text,int TstNum,StrTyp Typ)
{
   DWORD BgnTim,EndTim;

   BgnTim = GetTickCount();

   for (int i = 0; i < RepCnt; ++i)
   {
      int Pos = 0;

      T Word;

      while (Pos >=0)
      {
         Word = GetNextWord(Text,Pos);
         if (Pos > 0)
            Word[0] = toupper(Word[0]);
      }
   }

   EndTim = GetTickCount();
   Results[UprTstTyp][Typ][TstNum] = EndTim-BgnTim;
};
////
//      This procedure searches for a specified character within a Q-string.  The search begins   //
// at the specified offset and continues toward the end of the string.  The procedure returns the //
// offset of the first encountered character that matches the search character, or -1 if the      //
// search character is not found.  The string being searched is not affected by this procedure.   //
//                                                                                                //
//      This procedure is identical to the Scn2End() member procedure, except that if there is no //
// match, it returns -1.                                                                          //
template <class T>
int                                              // Zero-relative offset of the of the first      //
                                                 // occurrence of the search character found.  -1 //
                                                 // if the search character was not found.        //
Scn(const T &Str,
    char     Chr,                                // Character to search for.                      //
    int      BgnOff)                             // Default = 0.  Zero relative offset of the     //
                                                 // search starting position.  0 to start at the  //
                                                 // beginning of the string.                      //
{
   const int   Len = Str.length();
   const char *CP  = Str.data() + BgnOff;        // -> current character.                         //

   for (int i = BgnOff; i < Len; ++i)            // For each starting position.                   //
      if (*CP++ == Chr)                          // If search character is at position then,      //
         return(i);                              // Return position.                              //
   return -1;                                    // Return -1 to indicate no match.               //
}
////
//      This procedure searches for the first character within a Q-string that is not a specified //
// character.  The search begins at the specified offset and continues toward the end of the      //
// string.  The procedure returns the offset of the first encountered character that doesn't      //
// match the search character, or -1 if no character is found.  The string being searched is not  //
// affected by this procedure.                                                                    //
//                                                                                                //
//      This procedure is identical to the Skp2End() member procedure, except that if there is no //
// match, it returns -1.                                                                          //
template <class T>
int                                              // Zero-relative offset of the of the first      //
                                                 // occurrence of a non-search character found.   //
                                                 // -1 if a non-search character was not found.   //
Skp(const T &Str,
    char     Chr,                                // Character to search for.  (Skip over.)        //
    int      BgnOff)                             // Default = 0.  Zero relative offset of the     //
                                                 // search starting position.  0 to start at the  //
                                                 // beginning of the string.                      //
{
   const int   Len = Str.length();
   const char *CP  = Str.data() + BgnOff;        // -> current character.                         //

   for (int i = BgnOff; i < Len; ++i)            // For each starting position.                   //
      if (*CP++ != Chr)                          // If search character is not at position then,  //
         return(i);                              // Return position.                              //
   return -1;                                    // Return -1 to indicate no non-search character //
                                                 // was found.                                    //
}
////
//      This procedure searches for a specified character within a Q-string.  The search begins   //
// at the specified offset and continues toward the end of the string.  The procedure returns the //
// offset of the first encountered character that matches the search character, or the length of  //
// the string (an offset just past the end of the string) if the search character is not found.   //
// The string being searched is not affected by this procedure.                                   //
//                                                                                                //
//      This procedure is identical to the Scn() member procedure, except that if there is no     //
// match, it returns an offset that is past the end of the string.  This can be used to simplify  //
// some string processing algorithms.                                                             //
template <class T>
int                                              // Zero-relative offset of the of the first      //
                                                 // occurrence of the search character found.     //
                                                 // The length of the string if the search        //
                                                 // character was not found.                      //
Scn2End(const T &Str,
        char     Chr,                            // Character to search for.                      //
        int      BgnOff)                         // Default = 0.  Zero relative offset of the     //
                                                 // search starting position.  0 to start at the  //
                                                 // beginning of the string.                      //
{
   int i = Scn(Str,Chr,BgnOff);                  // Scan for character.                           //

   if (i == -1)                                  // If there was no match.                        //
      i = Str.length();                          // Get offset past end of string.                //
   return i;                                     // Return offset.                                //
}
////
//      This procedure searches for the first character within a Q-string that is not a specified //
// character.  The search begins at the specified offset and continues toward the end of the      //
// string.  The procedure returns the offset of the first encountered character that doesn't      //
// match the search character, or the length of the string (an offset just past the end of the    //
// string) if no character is found.  The string being searched is not affected by this           //
// procedure.                                                                                     //
//                                                                                                //
//      This procedure is identical to the Skp() member procedure, except that if there is no     //
// match, it returns an offset that is past the end of the string.  This can be used to simplify  //
// some string processing algorithms.                                                             //
template <class T>
int                                              // Zero-relative offset of the of the first      //
                                                 // occurrence of a non-search character found.   //
                                                 // The length of the string if a non-search      //
                                                 // character was not found.                      //
Skp2End(const T &Str,
        char     Chr,                            // Character to search for.  (Skip over.)        //
        int      BgnOff)                         // Default = 0.  Zero relative offset of the     //
                                                 // search starting position.  0 to start at the  //
                                                 // beginning of the string.                      //
{
   int i = Skp(Str,Chr,BgnOff);                  // Skip character.                               //

   if (i == -1)                                  // If there was no match.                        //
      i = Str.length();                          // Get offset past end of string.                //
   return i;                                     // Return offset.                                //
}
////
//      This procedure searches for a specified character within a Q-string.  The search begins   //
// at the specified offset and continues toward the start of the string.  The procedure returns   //
// the offset of the first encountered character that matches the search character, or -1 if the  //
// search character is not found.  The string being searched is not affected by this procedure.   //
template <class T>
int                                              // Zero-relative offset of the of the first      //
                                                 // occurrence of the search character found.  -1 //
                                                 // if the search character was not found.        //
RScn(const T &Str,
     char     Chr,                               // Character to search for.                      //
     int      BgnOff)                            // Default = -2.  Zero relative offset of the    //
                                                 // search starting position.  -2 to start at the //
                                                 // end of the string.                            //
{
   const int Len = Str.length();

   if (BgnOff == -2 || BgnOff >= Len)            // If search should start with end of string.    //
      BgnOff = Len - 1;                          // Get index of last character in the string.    //

   const char *CP = Str.data() + BgnOff;         // -> current chracter.                          //

   for (int i = BgnOff; i >= 0; --i)             // For each starting position.                   //
      if (*CP-- == Chr)                          // If search character is at position then,      //
         return(i);                              // Return position.                              //
   return -1;                                    // Return -1 to indicate no match.               //
}
////
template <class T>
void RunChrTst(const T &Text,int TstNum,StrTyp Typ)
{
   DWORD BgnTim,EndTim;

   BgnTim = GetTickCount();

   for (int i = 0; i < RepCnt; ++i)
   {
      for (int i = 0; i < Text.length(); ++i)
         Text[i];
   }

   EndTim = GetTickCount();
   Results[ChrTstTyp][Typ][TstNum] = EndTim-BgnTim;
};
template <class T>
T                                                // Next line.                                    //
GetNxtLin(const T &TxtStr,                       // String containing the text to be wrapped.     //
          int      WrpWdt,                       // Width of the wrapping area.                   //
          int     &BgnOff)                       // Offset to start of line to be returned.  The  //
                                                 // procedure will return the offset of the new   //
                                                 // next line or -1 when there are no more lines. //
                                                 // The way this is usually used is that an       //
                                                 // integer is initialized to 0 before the first  //
                                                 // call to GetNxtLin(), then that integer is     //
                                                 // just passed to each call to specify the       //
                                                 // current offset and get the new offset.  After //
                                                 // each call, the offset is checked, if it is    //
                                                 // -1, the end of the string has been reached    //
                                                 // and the string returned is ignored.           //
{
   if (BgnOff >= TxtStr.length())                // If past end of text string, then              //
   {
      BgnOff = -1;                               // Indicate there is no next line.               //
      return T();                                // Indicate there is no next line.               //
   }

   T   Txt    = TxtStr.substr(BgnOff,TxtStr.length()); // Remaining text.                         //
   int RetOff = Scn2End(Txt,'\n',0);             // Offset of next return, string length if none. //
   int RetLen = 1;                               // Length of the "return expression".            //

   if (Txt[RetOff-1] == '\r')                    // If there is a CR before the LF, then          //
   {
      RetLen = 2;                                // Indicate the return length is 2.              //
      --RetOff;                                  // Indicate return starts one character earlier. //
   }

   const int LinLen = RetOff;                    // Line length.  Same as RetOff.                 //

   Txt = Txt.substr(0,RetOff);                   // Trim up to return, or take whole string.      //

   int  BrkOff = WrpWdt;                         // Offset to character to break line at.         //
                                                 // This character will be kept and next won't.   //

   if (BrkOff >= LinLen)                         // If the whole line fits, then                  //
   {
      BrkOff = LinLen-1;                         // Indicate the whole line should be used.       //
   }
   else                                          // Otherwise, if line must be broken, then       //
   {
      if (Txt[BrkOff] == ' ')                    // If break is at a space, then                  //
      {
          BrkOff = Skp2End(Txt,' ',BrkOff) - 1;  // Scan past remaining spaces.                   //
      }
      else                                       // Otherwise, if break is at a character, then   //
      {
         int SpcOff = RScn(Txt,' ',BrkOff);      // Offset of last space before break point.      //

         if (SpcOff != -1)                       // If a space ws found.                          //
            BrkOff = SpcOff;                     // Break at space found.                         //
         else if (BrkOff > 0)                    // Otherwise, if no space fnd. and there's room, //
            --BrkOff;                            // Break before current break character (it      //
                                                 // doesn't fit).                                 //
      }
   }
   Txt = Txt.substr(0,BrkOff+1);                 // Calculate the length to keep.                 //

   if (BrkOff + 1 == RetOff)                     // if breaking at the return, then               //
      BrkOff += RetLen;                          // Adjust to skip past return.                   //
   BgnOff += BrkOff + 1;                         // Calculate next starting offset.               //
   return Txt;
}
////
template <class T>
void RunWrpTst(const T &Text,int TstNum,StrTyp Typ)
{
   DWORD BgnTim,EndTim;

   BgnTim = GetTickCount();

   for (int i = 0; i < RepCnt; ++i)
   {
      int CurOff  = 0;                           // Current offset into the string.               //
      T   WrpTxt;                                // Wrapped string.                               //

      while (true)                               // Repeat until done.                            //
      {
         T CurLin = GetNxtLin(Text,80,CurOff);   // Get the current line.                         //

         if (CurOff == -1)                       // If we are done, then                          //
            break;                               // Abort wrapping.                               //
         if (WrpTxt.length())                    // If this is not the first line.                //
            WrpTxt += '\n';                      // Start a new line.                             //
         WrpTxt += CurLin;                       // Add the line to the array.                    //
      }
   }

   EndTim = GetTickCount();
   Results[WrpTstTyp][Typ][TstNum] = EndTim-BgnTim;
};
////
void Test(ostream &OutPut)
{
 const char *Text =
"                      Step 1"
"1            package  Lorna Doone Cookies -- crumbled"
"   1/3           cup  sliced almonds -- crumbled"
"2        Tablespoons  butter -- softened"
"                      Step 2"
"2             pounds  cream cheese -- softened"
"7                     eggs"
"2               cups  sugar"
"2        Tablespoons  almond extract"
"2          teaspoons  vanilla extract"
"2               cups  sour cream"
"                      Step 4"
"2        Tablespoons  amaretto"
"1.   Grease bottom and sides of a 10 inch springform pan."
"     Mix crumbled cookies, sliced almonds, and 2 T softened butter until thoroughly mixed and crumbly."
"     Press into the bottom (and sides, if desired) of the springform pan."
"2.   Mix cream cheese with an electric mixer until it is the consistency of whipped cream."
"     In a medium-size bowl, combine the eggs, sugar, almond extract, and vanilla extract."
"     Pour into the cream cheese and mix until smooth."
"     Beat in sour cream."
"3.   Pour into prepared springform pan."
"     Bake at 350F until set, but still quite moist in the center, about 1 hour to 1 hour and 20 minutes."
"4.   Allow cake to cool completely, then pour the amaretto over the entire top of the cake."
"     Cover and refrigerate for at least 8 hours before slicing.";
   DWORD OptTim= 0;
   DWORD RefTim= 0;
   DWORD STLTim= 0;
   int TstNum;

   for (TstNum = 0; TstNum < TstCnt; ++TstNum)
   {
      cout << TstNum;
      RunRecTst(OptStr(Text),TstNum,OptStrTyp);
      RunRecTst(RefStr(Text),TstNum,RefStrTyp);
      RunRecTst(string(Text),TstNum,STLStrTyp);
   }
   cout << endl;
   for (TstNum = 0; TstNum < TstCnt; ++TstNum)
   {
      cout << TstNum;
      RunRevTst(OptStr(Text),TstNum,OptStrTyp);
      RunRevTst(RefStr(Text),TstNum,RefStrTyp);
      RunRevTst(string(Text),TstNum,STLStrTyp);
   }
   cout << endl;
   for (TstNum = 0; TstNum < TstCnt; ++TstNum)
   {
      cout << TstNum;
      RunChrTst(OptStr(Text),TstNum,OptStrTyp);
      RunChrTst(RefStr(Text),TstNum,RefStrTyp);
      RunChrTst(string(Text),TstNum,STLStrTyp);
   }
   cout << endl;
   for (TstNum = 0; TstNum < TstCnt; ++TstNum)
   {
      cout << TstNum;
      RunUprTst(OptStr(Text),TstNum,OptStrTyp);
      RunUprTst(RefStr(Text),TstNum,RefStrTyp);
      RunUprTst(string(Text),TstNum,STLStrTyp);
   }
   cout << endl;

   for (TstNum = 0; TstNum < TstCnt; ++TstNum)
   {
      cout << TstNum;
      RunWrpTst(OptStr(Text),TstNum,OptStrTyp);
      RunWrpTst(RefStr(Text),TstNum,RefStrTyp);
      RunWrpTst(string(Text),TstNum,STLStrTyp);
   }
   cout << endl;

   OutPut << "Repetitions: " << RepCnt << endl;
   for (int TstTyp = 0; TstTyp < MaxTstTyp; ++TstTyp)
   {
      OutPut << TstTypNamLst[TstTyp] << endl;
      for (int StrTyp = 0; StrTyp < MaxStrTyp; ++StrTyp)
      {
         OutPut << StrTypNamLst[StrTyp];
         for (TstNum = 0; TstNum < TstCnt; ++TstNum)
            OutPut << '\t' << Results[TstTyp][StrTyp][TstNum];
         OutPut << endl;
      }
   }
}
////
int main()
{
   string FilNam;

   cout << "Output file: ";
   getline(cin,FilNam);

   ofstream Out(FilNam.c_str());
   cout << "begin" << endl;
   Test(Out);

   return 0;
}